# SWMODD

Software Module(`swmodd`) Daemon manages the lxc and OCI based containers(using liblxc and crun) and provides the Backend for Software Modules as defined in [USP](https://usp.technology/specification/index.pdf) and [CWMP](https://www.broadband-forum.org/technical/download/TR-069.pdf).

TR181 data model parameters for [Device.SoftwareModules.](https://usp-data-models.broadband-forum.org/tr-181-2-15-1-usp.html#D.Device:2.Device.SoftwareModules.) exposed using bbf [DotSO Plugin](https://dev.iopsys.eu/iopsys/bbf/-/blob/devel/README.md).

It is written in C programming language and depends on a number of libraries of OpenWrt for building and running.

## Good to Know
It currently has support for LXC containers and OCI specification based application containers.

Application containers has all the dependencies internally resolved, so it does not required further modification inside the container. An OCI based application container can be installed using `du_install` ubus method. After successful installation, the OCI container will be visible in `du_list` and `eu_list` ubus method.

LXC based containers might requires further addition of packages on top of the base container, currently procd based lxc containers along with opkg only supported. Addition of new service/packages done using `opkg`, so all the `opkg` limitations can be considered as limitation for `swmodd` as well with respect to listing/installing/upgrading packages/services.

>NOTE:
- Installation of lxc base container is out of scope of swmodd
- ExecEnv.1 is Virtual environment used for installation of application containers
- Installation of `opkg` based DU not supported on ExecEnv.1
- Application containers support installation from tar ball or container registry
- In lxc ExecEnv, only addon package/services shall be visible in `du_list/eu_list`

## Concepts and Workflow
`SWMODD` usages `lxc` library to interact with the lxc containers and `opkg` system utilities to manage the services running inside that lxc container, for OCI application containers it usages `crun` to maintain the life cycle of the container.

Installation of application container from image registry work as a add-on feature, and is dependent on external modules `skopeo` and `umoci`.

`swmodd` is used to manage the software modules and exposes the functionality over ubus, whereas `libswmodd.so` is `bbf` plugin which exposes the SoftwareModules functionality over TR181 using `libbbf_api`.

## swmodd uci
Below is the uci for swmodd

```bash
config globals 'globals'
	option enabled '1'
	option lan_bridge 'br-lan'
	option oci_bundle_root '/data/crun'
	option lxc_bundle_root '/data/lxc'
```

For installation of oci based application containers, bundle path need to defined in above uci in 'oci_bundle_root' option.
For installation of lxc based containers, bundle path need to defined in above uci in 'lxc_bundle_root' option.
It is suggested to configure a non-volatile secondary memory location in this options. Also 'lan_bridge' need to point to a valid bridge where application containers network namespace shall be attached.

## swmodd uBus

`swmodd` must be launched on startup after `uspd`. It exposes the SoftwareModules datamodel objectes over ubus. The `swmodd` registers `swmodules` namespaces with ubus, that has the shown below functionalities:

```bash
root@iopsys:~# ubus -v list swmodules
'swmodules' @19113d4c
        "ee_list":{}
        "reload":{}
        "du_list":{"eeid":"Integer","ee_name":"String"}
        "eu_list":{"eeid":"Integer","ee_name":"String"}
        "du_install":{"eeid":"Integer","ee_name":"String","uuid":"String","url":"String","username":"String","password":"String"}
        "du_update":{"eeid":"Integer","ee_name":"String","uuid":"String","url":"String","username":"String","password":"String"}
        "du_uninstall":{"eeid":"Integer","ee_name":"String","du_name":"String"}
        "eu_set_state":{"eeid":"Integer","ee_name":"String","eu_name":"String","state":"Boolean"}
        "ee_set_state":{"eeid":"Integer","ee_name":"String","state":"String"}
root@iopsys:~#
```

For more info on the `swmodules` ubus schema see [link](./docs/api/swmodules.md) or [raw schema](./schemas/ubus/swmodules.json)

## swmodd lxc templates
swmodd also includes LXC templates to easily create LXC based containers. This template shall used only for development and test purpose.

#### To create a new container using iopsys template:

`lxc-create -n <container_name> -t iopsys`
example:
```bash
root@iopsys:~# lxc-create -n test -t iopsys
root@iopsys:~#
```

### How to deploy application containers
`swmodd` support deployment of application containers which support OCI image and runtime formats using CRUN.

#### How to create an application container tar
To convert an docker image to OCI application container, we need to use `skopeo` and `umoci` utilities.

In below exaple [httpd](https://hub.docker.com/_/httpd) docker container is used
- Create an empty directory in local laptop/desktop
```bash
$ mkdir lcm
$ cd lcm
```
- Copy the docker image using skopeo, make sure to use correct tag as per the target architecture
```bash
$ skopeo copy docker://httpd:latest oci:httpd_copy:latest
```
- Unpack the downloaded image in OCI format using umoci
```bash
$ sudo -i
# cd <working_dir_path>
# umoci unpack --image httpd_copy:latest httpd
```
- Move to the directory containing config.json and rootfs and create a tar
```bash
# cd httpd
# tar cf ../httpd.tar *
```

- Use this tar in `install_du` ubus command to deploy the application container

##### Sulu application container
An example makefile to generate sulu application container tar ball can be accessed [here](./examples/sulu/README.md)

### swmodd ubus examples
The outputs shown below are just an example, it can vary on each system. Long outputs are truncated to beautify the document.

#### List down the available execution environments
```bash
root@iopsys:~# ubus call swmodules ee_list
{
        "environment": [
                {
                        "ee_name": "OpenWRT_Linux",
                        "eeid": 1,
                        "status": "Up",
                        "pause": 0,
                        "autoboot": 1,
                        "type": "Virtual_EE",
                        "vendor": "iopsys-eca940fff680",
                        "version": "4.19.235",
                        "allocated_disk_space": 27632,
                        "available_disk_space": 26716,
                        "allocated_memory": 1014988,
                        "available_memory": 476840,
                        "parent_ee_ref": 0
                }
        ]
}
root@iopsys:~#
```
#### Install new application container as deployment unit
This requires support to pull images from container registry.

```bash
root@iopsys:~# ubus call swmodules du_install '{"url":"docker://httpd:latest"}'
{
        "status": true
}
root@iopsys:~#
```
#### List available deployment units/installed services
```bash
root@iopsys:~# ubus call swmodules du_list
{
        "deployment_unit": [
                {
                        "du_name": "httpd",
                        "ee_name": "OpenWRT_Linux",
                        "eeid": 1,
                        "uuid": "347dd4bf-6b51-4d62-be61-77c8ded98698",
                        "duid": "w0t0000",
                        "url": "docker://httpd:latest",
                        "version": "",
                        "config": "",
                        "description": "",
                        "du_status": "Installed",
                        "eu_name": "httpd",
                        "vendor": ""
                }
        ]
}
root@iopsys:~#
```

#### List execution units/running services
```bash
root@iopsys:~# ubus call swmodules eu_list
{
        "execution_unit": [
                {
                        "eu_name": "httpd",
                        "command": "",
                        "state": "Active",
                        "config": "",
                        "version": "4.19.235",
                        "description": "This is a CRUN container",
                        "ee_name": "OpenWRT_Linux",
                        "eeid": 1,
                        "euid": "22728",
                        "disk_space": 91840,
                        "memory_space": 1014988,
                        "vendor": "umoci-default",
                        "req_state": "Active",
                        "autostart": true,
                        "du_name": "httpd"
                }
        ]
}
```

## Dependencies
`swmodd` compile time and run time dependencies.

### Build-Time Dependencies

To successfully build `swmodd`, the following libraries are needed:

| Dependency  | Link                                          | License                                |
| ----------- | --------------------------------------------- | -------------------------------------- |
| libuci      | https://git.openwrt.org/project/uci.git       | LGPL 2.1                               |
| libubox     | https://git.openwrt.org/project/libubox.git   | BSD                                    |
| ubus        | https://git.openwrt.org/project/ubus.git      | LGPL 2.1                               |
| liblxc      | https://linuxcontainers.org/downloads/lxc/    | LGPL-2.1-or-later BSD-2-Clause GPL-2.0 |
| opkg        | https://git.openwrt.org/project/opkg-lede.git | GPL-2.0                                |
| libuuid     | util-linux                                    | GPL-2.0                                |
| libcurl     | https://dl.uxnr.de/mirror/curl                | MIT                                    |


### Run-Time Dependencies

In order to run the `swmodd`, following dependencies are needed to be running/available before `swmodd`.

| Dependency |                   Link                        | License    |
| ---------- | --------------------------------------------- | ---------- |
| ubusd      | https://git.openwrt.org/project/ubus.git      | LGPL 2.1   |
| bbf        | https://dev.iopsys.eu/iopsys/bbf.git          | LGPL 2.1   |
| opkg       | https://git.openwrt.org/project/opkg-lede.git | GPL-2.0    |
| uspd       | https://dev.iopsys.eu/iopsys/uspd.git         | GPL-2.0    |
| libcurl    | https://dl.uxnr.de/mirror/curl                | MIT        |
| umoci      | https://github.com/opencontainers/umoci.git   | Apache-2.0 |
| Skopeo     | https://github.com/containers/skopeo.git      | Apache-2.0 |

Apart from these dependencies, LXC-4.0 > also requires cgroup functionalities to be mounted separately, to do so, `swmodd-cgroup` package is used.
> `uspd` along with `bbf` is used to expose the SoftwareModule datamodel objects.
> umoci and skopeo required to pull images from container registries

