/*
 * swmod_uci.h: SWMOD deamon
 *
 * Copyright (C) 2020 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Amin Ben Ramdhane <amin.benramdhane@pivasoftware.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#ifndef __SWMOD_UCI_H
#define __SWMOD_UCI_H

#include <uci.h>

int swmod_uci_init(const char *conf_path);
int swmod_uci_fini(const char *package_name);

struct uci_section *swmod_uci_add_section(const char *package, const char *section_type, bool rename);
int swmod_uci_delete_by_section(struct uci_section *section, const char *option, const char *value);

char *swmod_uci_get_value_by_section(struct uci_section *section, const char *option);
char *swmod_uci_set_value_by_section(struct uci_section *section, const char *option, const char *value);
char *swmod_uci_set_value_by_section_list(struct uci_section *section, const char *option, const char *value);

struct uci_section *swmod_uci_walk_section(const char *package, const char *section_type, struct uci_section *prev_section);

bool check_section_exist_by_option(const char *section, const char *section_type, const char *option, const char *val_check, struct uci_section **s);

bool uci_str_to_bool(const char *value);
int swmod_uci_commit(char *package);

#define swmod_uci_foreach_section(package, section_type, section) \
	for (section = swmod_uci_walk_section(package, section_type, NULL); \
		section != NULL; \
		section = swmod_uci_walk_section(package, section_type, section))

#define swmod_uci_foreach_section_safe(package, section_type, _tmp, section) \
	for(section = swmod_uci_walk_section(package, section_type, NULL), \
		_tmp = (section) ? swmod_uci_walk_section(package, section_type, section) : NULL; \
		section != NULL; \
		section = _tmp, _tmp = (section) ? swmod_uci_walk_section(package, section_type, section) : NULL)

#endif //__SWMOD_UCI_H
